/*
 * Copyright (c) 2004, PostgreSQL Global Development Group
 * See the LICENSE file in the project root for more information.
 */
// Copyright (c) 2004, Open Cloud Limited.

package org.postgresql.core.v3;

import java.lang.ref.PhantomReference;
import java.nio.charset.StandardCharsets;
import org.postgresql.core.ResultCursor;

/**
 * V3 ResultCursor implementation in terms of backend Portals. This holds the state of a single
 * Portal. We use a PhantomReference managed by our caller to handle resource cleanup.
 *
 * @author Oliver Jowett (oliver@opencloud.com)
 */
class Portal implements ResultCursor {
    private final SimpleQuery query;
    private final String portalName;
    private final byte[] encodedName;
    private PhantomReference<?> cleanupRef;

    Portal(SimpleQuery query, String portalName) {
        this.query = query;
        this.portalName = portalName;
        this.encodedName = portalName.getBytes(StandardCharsets.UTF_8);
    }

    @Override
    public void close() {
        PhantomReference<?> cleanupRef = this.cleanupRef;
        if (cleanupRef != null) {
            cleanupRef.clear();
            cleanupRef.enqueue();
            this.cleanupRef = null;
        }
    }

    String getPortalName() {
        return portalName;
    }

    // Holding on to a reference to the generating query has
    // the nice side-effect that while this Portal is referenced,
    // so is the SimpleQuery, so the underlying statement won't
    // be closed while the portal is open (the backend closes
    // all open portals when the statement is closed)

    byte[] getEncodedPortalName() {
        return encodedName;
    }

    SimpleQuery getQuery() {
        return query;
    }

    void setCleanupRef(PhantomReference<?> cleanupRef) {
        this.cleanupRef = cleanupRef;
    }

    @Override
    public String toString() {
        return portalName;
    }
}
